import json
from typing import Any, List
from mlserver.codecs import StringCodec, Base64Codec, register_input_codec, register_request_codec
from mlserver.codecs.utils import SingleInputRequestCodec
from mlserver.types import RequestInput, Parameters, ResponseOutput
from mlserver.errors import InferenceError
from .lib import unpack
import base64
import gzip

def _decompress_base64_gzip_json(body: str) -> str:
    return gzip.decompress(base64.b64decode(body)).decode('utf-8')

def _compress_base64_gzip_json(data: Any) -> str:
    return base64.b64encode(gzip.compress(json.dumps(data).encode('utf-8'))).decode('utf-8')

@register_input_codec
class JSONGzippedBase64Codec(Base64Codec):
    ContentType = "base64_gzip_json"

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return False

    @classmethod
    def encode_output(cls, name: str, payload: List[Any], **kwargs) -> ResponseOutput:
        shape = [len(payload), 1]
        print(list(map(_compress_base64_gzip_json, payload)))
        return ResponseOutput(
            name=name,
            datatype="BYTES",
            shape=shape,
            data=list(map(_compress_base64_gzip_json, payload)),
            parameters=Parameters(content_type=cls.ContentType),
        )

    @classmethod
    def encode_input(cls, name: str, payload: List[Any], **kwargs) -> RequestInput:
        output = cls.encode_output(name=name, payload=payload)

        return RequestInput(
            name=output.name,
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
            parameters=Parameters(content_type=cls.ContentType),
        )

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> Any:  # type: ignore
        try:
            packed = request_input.data.__root__
            unpacked = list(map(_decompress_base64_gzip_json, unpack(packed)))
            print(unpacked)
            return json.loads("[" + ",".join([sentence for sentence in unpacked]) + "]")
        except Exception as e:
            # There are a few things that can go wrong here, e.g. less than 2-D
            # in the array), or input data not compatible with a numpy array
            raise InferenceError("Invalid input to JSONGzippedBase64") from e

@register_request_codec
class JSONGzippedBase64RequestCodec(SingleInputRequestCodec):
    InputCodec = JSONGzippedBase64Codec
    ContentType = JSONGzippedBase64Codec.ContentType

@register_input_codec
class JSONCodec(StringCodec):
    ContentType = "json"

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return False

    @classmethod
    def decode_input(cls, request_input: RequestInput) -> Any:  # type: ignore
        try:
            unpacked = super().decode_input(request_input)
            # out = []
            # for k in unpacked:
            #     out.append(json.loads(k))
            # return out
            return json.loads("[" + ",".join([sentence for sentence in unpacked]) + "]")

        except Exception as e:
            # There are a few things that can go wrong here, e.g. less than 2-D
            # in the array), or input data not compatible with a numpy array
            raise InferenceError("Invalid input to JSON") from e


@register_request_codec
class JSONRequestCodec(SingleInputRequestCodec):
    InputCodec = JSONCodec
    ContentType = JSONCodec.ContentType


from mlserver import MLModel
from mlserver.types import InferenceRequest, InferenceResponse, ResponseOutput
from mlserver.codecs import NumpyCodec
from mlserver.errors import InferenceError, ModelNotFound
import os
from openvino.runtime import Core
import onnx
from mlserver.utils import get_model_uri
from typing import List
from .transformer import Transformer
import time
from mlserver.logging import get_logger
# Dont remove this, This line load and auto register custom requests codecs
from .codecs import NumpyGzipCodec, JSONGzippedBase64Codec

logger = get_logger()

WELLKNOWN_MODEL_FILENAMES = ["model.xml", "model.onnx"]

class OpenvinoRuntime(MLModel):

  async def load(self) -> bool:

    model_uri = await get_model_uri(
      self._settings, wellknown_filenames=WELLKNOWN_MODEL_FILENAMES
    )

    # Should save model path before convert from onnx
    self.model_dir = os.path.dirname(model_uri)

    # convert onnx file to openvino format
    if model_uri.endswith('.onnx'):
      model_uri = self.convert_onnx(model_uri)


    # Start model openvino
    self.core = Core()

    if not os.path.exists(model_uri):
      raise ModelNotFound("Cant find model file")

    self._model = self.core.read_model(model=model_uri)
    self.compiled_model_int8 = self.core.compile_model(model=self._model, device_name="CPU")
    self.model_inputs = [list(inp.names)[0] for inp in self.compiled_model_int8.inputs]
    self.model_outputs = [list(inp.names)[0] for inp in self.compiled_model_int8.outputs]
    self.transformers: List[Transformer] = self._load_transformers()
    self.has_transformers = len(self.transformers) > 0

    return await super().load()

  def _load_transformers(self) -> List[Transformer]:
    if 'transform' not in self.settings.parameters.extra or len(self.settings.parameters.extra['transform']) == 0:
      return []

    transformers = []
    for pipeline in self.settings.parameters.extra['transform']:
      transformers.append(
        Transformer(
          os.path.join(self.model_dir, pipeline['pipeline_file_path']),
          pipeline['name'],
          pipeline['input_index']
        )
      )

    return transformers

  async def predict(self, payload: InferenceRequest) -> InferenceResponse:
    payload = self._check_request(payload)
    outputs = self._get_model_outputs(payload)

    return InferenceResponse(
      id=payload.id,
      model_name=self.name,
      model_version=self.version,
      outputs=outputs,
    )

  def _get_model_outputs(self, payload: InferenceRequest) -> List[ResponseOutput]:
    for request_output in payload.outputs:
      try:

        start_time = time.time()
        X = self._preprocessing(
          [self.decode(inp, default_codec=NumpyCodec) for inp in payload.inputs]
        )
        logger.debug("Preprocessing elapsed time: " + str(time.time() - start_time))

        outputs = []
        output_layer_index = self.model_outputs.index(request_output.name)

        start_time = time.time()
        y = self.compiled_model_int8(inputs=X)[self.compiled_model_int8.outputs[output_layer_index]]
        logger.debug("Predict elapsed time: " + str(time.time() - start_time))

        output = self.encode(y, request_output, default_codec=NumpyCodec)
        outputs.append(output)
      except ValueError:
        raise InferenceError(f"Cant find output with name {request_output.name}")

    return outputs

  def _preprocessing(self, X: List) -> List:
    if self.has_transformers:
      return self._execute_transform_pipeline(X)

    return X

  def _execute_transform_pipeline(self, X:List) -> List:
    for tranformer_pipeline in self.transformers:
      X = tranformer_pipeline.transform(X)

    return X

  def _check_request(self, payload: InferenceRequest) -> InferenceResponse:
    for inp in payload.inputs:
      if inp.name not in self.model_inputs:
        raise InferenceError(f"No {inp.name} input in model")

    return payload

  @classmethod
  def convert_onnx(cls, file_path, output_path='/tmp/openvino'):
    '''
    Convert onnx file to openvino format
    :param file_path: str path to onnx file
    :return: None
    '''
    model_onnx = onnx.load(file_path)
    inputs_names = []
    inputs_dim = []
    for inp in model_onnx.graph.input:
      inputs_names.append(inp.name)
      dims = [str(d.dim_value) for d in inp.type.tensor_type.shape.dim]
      dims[0] = "-1"

      inputs_dim.append('[' + ",".join(dims) + ']')

    # Create dir in tmp
    if not os.path.exists(output_path):
      os.makedirs(output_path)

    # Command
    command = f"mo --input_model \"{file_path}\" --input_shape={','.join(inputs_dim)} --input={','.join(inputs_names)} --output output --output_dir \"{output_path}\" --model_name \"model\""
    os.system(command)

    return os.path.join(output_path, 'model.xml')




from typing import Any, List, Union
from mlserver.codecs import InputCodec, register_input_codec, register_request_codec
from mlserver.codecs.numpy import to_dtype, to_datatype
from mlserver.codecs.lists import ListElement
from mlserver.codecs.utils import inject_batch_dimension, SingleInputRequestCodec
from mlserver.types import RequestInput, ResponseOutput, Parameters
from mlserver.errors import InferenceError
from .lib import unpack
import base64
import gzip
import numpy as np
import binascii

_Base64StrCodec = "ascii"

def _ensure_bytes(elem: ListElement) -> bytes:
    if isinstance(elem, str):
        return elem.encode(_Base64StrCodec)
    return elem

def _encode_base64(elem: ListElement, use_bytes: bool) -> Union[bytes, str]:
    as_bytes = _ensure_bytes(elem)
    b64_encoded = base64.b64encode(as_bytes)
    if use_bytes:
        return b64_encoded

    return b64_encoded.decode(_Base64StrCodec)


def _decode_base64(elem: ListElement) -> bytes:
    as_bytes = _ensure_bytes(elem)

    # Check that the input is valid base64.
    # Otherwise, convert into base64.
    try:
        return base64.b64decode(as_bytes, validate=True)
    except binascii.Error:
        return as_bytes

def _encode_data(data: np.ndarray) -> list:
    return [_encode_base64(gzip.compress(data.tobytes()), False)]

def _decompress_numpy(body: str) -> bytes:
    return gzip.decompress(base64.b64decode(body))

@register_input_codec
class NumpyGzipCodec(InputCodec):
    ContentType = "np_gzip"

    @classmethod
    def can_encode(cls, payload: Any) -> bool:
        return False

    @classmethod
    def decode_output(cls, response_output: ResponseOutput) -> np.ndarray:
        return cls.decode_input(response_output)  # type: ignore

    @classmethod
    def encode_output(cls, name: str, payload: np.ndarray, **kwargs) -> ResponseOutput:

        datatype = to_datatype(payload.dtype)
        shape = inject_batch_dimension(list(payload.shape))

        return ResponseOutput(
            name=name,
            datatype=datatype,
            shape=shape,
            data=_encode_data(payload),
        )

    @classmethod
    def encode_input(cls, name: str, payload: np.ndarray, **kwargs) -> RequestInput:
        output = cls.encode_output(name=name, payload=payload)

        return RequestInput(
            name=output.name,
            datatype=output.datatype,
            shape=output.shape,
            data=output.data,
            parameters=Parameters(content_type=cls.ContentType),
        )


    @classmethod
    def decode_input(cls, request_input: RequestInput) -> Any:  # type: ignore
        try:
            dtype = to_dtype(request_input)

            packed = request_input.data.__root__
            unpacked = map(_decompress_numpy, unpack(packed))
            decoded = [np.frombuffer(sentence, dtype) for sentence in unpacked]
            return np.concatenate(decoded).reshape(request_input.shape)

        except Exception as e:
            # There are a few things that can go wrong here, e.g. less than 2-D
            # in the array), or input data not compatible with a numpy array
            raise InferenceError(e) from e

@register_request_codec
class NumpyRequestCodec(SingleInputRequestCodec):
    InputCodec = NumpyGzipCodec
    ContentType = NumpyGzipCodec.ContentType


from sklearn.base import TransformerMixin

class DumpTransform(TransformerMixin):
    '''
    Convert text and prepare to Neural Network input
    '''

    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return X


from typing import List
import joblib
import cloudpickle
import pickle
import gzip
from mlserver.logging import get_logger
import time

logger = get_logger()

class Transformer:
    '''
    Transform input for specific input index
    '''
    def __init__(self, file_path: str, name: str, input_index: int):
        self._name = name
        self._file_path = file_path
        self._input_index = input_index
        self._pipeline = None

        self._load()

    def _load(self):
        logger.debug(f"Loading transformer {self._file_path}")
        if 'cloudpickle' in self._file_path:
            self._load_cloudpickle()
        elif self._file_path.endswith('pickle'):
            self._load_pickle()
        else:
            self._load_joblib()

    def _load_pickle(self):
        with open(self._file_path,"rb") as f:
            self._pipeline = pickle.load(f)

    def _load_cloudpickle(self):
        if self._file_path.endswith('.gz'):
            with gzip.open(self._file_path, 'rb') as f:
                self._pipeline = cloudpickle.load(f)
        else:
            with open(self._file_path, "rb") as f:
                self._pipeline = cloudpickle.load(f)

    def _load_joblib(self):
        self._pipeline = joblib.load(self._file_path)

    def transform(self, X: List) -> List:
        start_time = time.time()
        X[self._input_index] = self._pipeline.transform(X[self._input_index])
        logger.debug(f"Transform {self._name} elapsed {time.time() - start_time}")
        return X


from typing import Generator, Union, List
import gzip
import base64

PackElement = Union[bytes, str]
PackedPayload = Union[PackElement, List[PackElement]]

def _decompress_gzip_json(body):
    return gzip.decompress(body).decode('utf-8')

def _decompress_base64_gzip_json(body):
    return gzip.decompress(base64.b64decode(body)).decode('utf-8')

def unpack(packed: PackedPayload) -> Generator[PackElement, None, None]:
    if isinstance(packed, list):
        # If it's a list, assume list of strings
        yield from packed
    else:
        # If there is no shape, assume that it's a single element
        yield packed